import math
from typing import Dict, List, Optional, Tuple, Union

from picsellia_annotations.coco import Annotation as COCOAnnotation
from picsellia_annotations.coco import Category as COCOCategory
from picsellia_annotations.coco import COCOFile
from picsellia_annotations.coco import Image as COCOImage

from picsellia.exceptions import ResourceNotFoundError
from picsellia.sdk.asset import MultiAsset


class COCOFileBuilder:
    def __init__(self, enforced_ordered_labels: Optional[List[str]] = None):
        self.labels = {}
        self.enforced_ordered_labels = enforced_ordered_labels
        if enforced_ordered_labels:
            for k in range(len(enforced_ordered_labels)):
                self.labels[enforced_ordered_labels[k]] = k

        self.assets = {}
        self.coco_annotations = []
        self.coco_images = []
        self.coco_categories = []

    def load_coco_annotations(self, loaded_annotations: Dict[str, list]) -> List[str]:
        for asset_id, annotations in loaded_annotations.items():
            # Find image id in already seen assets, store image_id for each asset
            if asset_id in self.assets:  # pragma: no cover
                image_id = self.assets[asset_id]
            else:
                image_id = len(self.assets)
                self.assets[asset_id] = image_id

            self._build_coco_shapes(image_id, annotations)

        return list(self.assets.keys())

    def load_coco_images(self, loaded_assets: MultiAsset):
        for asset in loaded_assets:
            coco_image = COCOImage(
                id=self.assets[str(asset.id)],
                file_name=asset.filename,
                width=asset.width,
                height=asset.height,
            )
            self.coco_images.append(coco_image)

    def load_coco_categories(self, loaded_label_names: List[str]):
        for label_name in loaded_label_names:
            if label_name not in self.labels:
                continue
            category = COCOCategory(id=self.labels[label_name], name=label_name)
            self.coco_categories.append(category)

        # Return categories by order of id
        self.coco_categories = sorted(self.coco_categories, key=lambda cat: cat.id)

    def build(self) -> COCOFile:
        return COCOFile(
            categories=self.coco_categories,
            images=self.coco_images,
            annotations=self.coco_annotations,
        )

    def _build_coco_shapes(
        self,
        image_id: int,
        annotations: List,
    ):
        for annotation in annotations:
            if "classifications" in annotation and annotation["classifications"]:
                for classification in annotation["classifications"]:
                    category_id = self._retrieve_label(classification["label"])
                    coco_annotation = COCOAnnotation(
                        id=len(self.coco_annotations),
                        image_id=image_id,
                        category_id=category_id,
                        bbox=[],
                        segmentation=[],
                    )
                    self.coco_annotations.append(coco_annotation)

            if "rectangles" in annotation and annotation["rectangles"]:
                for rectangle in annotation["rectangles"]:
                    category_id = self._retrieve_label(rectangle["label"])
                    coco_annotation = COCOAnnotation(
                        id=len(self.coco_annotations),
                        image_id=image_id,
                        category_id=category_id,
                        bbox=(
                            rectangle["x"],
                            rectangle["y"],
                            rectangle["w"],
                            rectangle["h"],
                        ),
                        segmentation=[],
                    )
                    self.coco_annotations.append(coco_annotation)

            if "polygons" in annotation and annotation["polygons"]:
                for polygon in annotation["polygons"]:
                    if (
                        "polygon" not in polygon or not polygon["polygon"]
                    ):  # pragma: no cover
                        continue
                    category_id = self._retrieve_label(polygon["label"])
                    bbox, shape = COCOFileBuilder._compute_bbox_and_shape(
                        polygon["polygon"]
                    )
                    coco_annotation = COCOAnnotation(
                        id=len(self.coco_annotations),
                        image_id=image_id,
                        category_id=category_id,
                        bbox=bbox,
                        segmentation=[shape],  # COCO needs an array around polygons
                    )
                    self.coco_annotations.append(coco_annotation)

    def _retrieve_label(self, label: str):
        if label not in self.labels:
            if self.enforced_ordered_labels:
                raise ResourceNotFoundError(
                    f"{label} is used by an asset of this dataset but was not given in {self.enforced_ordered_labels}"
                )
            self.labels[label] = len(self.labels)
        return self.labels[label]

    @staticmethod
    def _compute_bbox_and_shape(
        polygon: List[List[Union[int, float]]]
    ) -> Tuple[
        Tuple[
            Union[int, float], Union[int, float], Union[int, float], Union[int, float]
        ],
        List[Union[int, float]],
    ]:
        x_min = math.inf
        x_max = -1
        y_min = math.inf
        y_max = -1

        segmentation = []
        for point in polygon:
            x = point[0]
            x_min = min(x, x_min)
            x_max = max(x, x_max)
            y = point[1]
            y_min = min(y, y_min)
            y_max = max(y, y_max)
            segmentation.append(x)
            segmentation.append(y)

        return (x_min, y_min, x_max - x_min, y_max - y_min), segmentation



from threading import Lock
from typing import List


class ErrorManager:
    def __init__(self):
        self._errors: List[Exception] = []
        self._lock: Lock = Lock()

    @property
    def errors(self):
        with self._lock:
            return self._errors

    def has_errors(self):
        with self._lock:
            return self._errors is not None and self._errors != []

    def clear(self):
        with self._lock:
            self._errors = []

    def append(self, error: Exception):
        with self._lock:
            self._errors.append(error)

    def __str__(self):
        with self._lock:
            if self._errors:
                return f"This ErrorManager has {len(self._errors)} errors."
            else:
                return "This ErrorManager has no error."



from enum import Enum


class StrEnum(Enum):
    def __repr__(self):
        return "<%s.%s>" % (self.__class__.__name__, self.name)

    @classmethod
    def values(cls):
        return list(map(lambda c: c.value, cls))

    @classmethod
    def validate(cls, token):
        if isinstance(token, cls):
            return token
        elif isinstance(token, str):
            try:
                return cls[token.upper()]
            except KeyError:
                try:
                    return cls(token.upper())
                except ValueError:
                    raise TypeError(f"Given string should be one of {cls.values()}.")
        else:  # pragma: no cover
            # Should not happen if beartype is set
            raise TypeError(
                f"Given string should be a string one of {cls.values()} or a {cls}"
            )


class ExperimentStatus(StrEnum):
    RUNNING = "RUNNING"
    CANCELED = "CANCELED"
    TERMINATED = "TERMINATED"
    FAILED = "FAILED"
    WAITING = "WAITING"
    SUCCESS = "SUCCESS"


class LogType(StrEnum):
    VALUE = "VALUE"
    IMAGE = "IMAGE"
    LINE = "LINE"
    TABLE = "TABLE"
    BAR = "BAR"
    HEATMAP = "HEATMAP"
    LABELMAP = "LABELMAP"
    EVALUATION = "EVALUATION"


class RunStatus(StrEnum):
    WAITING = "WAITING"
    RUNNING = "RUNNING"
    CANCELED = "CANCELED"
    SUCCESS = "SUCCESS"
    TERMINATED = "TERMINATED"
    FAILED = "FAILED"


class DataType(StrEnum):
    IMAGE = "IMAGE"
    VIDEO = "VIDEO"


class ObjectDataType(StrEnum):
    ARTIFACT = "ARTIFACT"
    DATA = "DATA"
    LOGGING = "LOGGING"
    LOG_IMAGE = "LOG_IMAGE"
    MODEL_FILE = "MODEL_FILE"
    MODEL_THUMB = "MODEL_THUMB"
    SCAN_FILE = "SCAN_FILE"


class AnnotationStatus(StrEnum):
    PENDING = "PENDING"
    SKIPPED = "SKIPPED"
    ACCEPTED = "ACCEPTED"
    REFUSED = "REFUSED"


class AnnotationFileType(StrEnum):
    PASCAL_VOC = "PASCAL_VOC"
    COCO = "COCO"


class JobRunStatus(StrEnum):
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"
    KILLED = "KILLED"


class JobStatus(StrEnum):
    RUNNING = "RUNNING"
    SUCCESS = "SUCCESS"
    TERMINATED = "TERMINATED"
    FAILED = "FAILED"


class Framework(StrEnum):
    TENSORFLOW = "TENSORFLOW"
    PYTORCH = "PYTORCH"
    ONNX = "ONNX"
    NOT_CONFIGURED = "NOT_CONFIGURED"


class InferenceType(StrEnum):
    NOT_CONFIGURED = "NOT_CONFIGURED"
    CLASSIFICATION = "CLASSIFICATION"
    OBJECT_DETECTION = "OBJECT_DETECTION"
    SEGMENTATION = "SEGMENTATION"
    POINT = "POINT"
    LINE = "LINE"
    MULTI = "MULTI"


class TagTarget(StrEnum):
    DATA = "DATA"
    ASSET = "ASSET"
    MODEL_VERSION = "MODEL_VERSION"
    MODEL = "MODEL"
    DATASET_VERSION = "DATASET_VERSION"
    DATASET = "DATASET"
    DEPLOYMENT = "DEPLOYMENT"
    PREDICTED_ASSET = "PREDICTED_ASSET"


class ImportAnnotationMode(StrEnum):
    REPLACE = "REPLACE"
    KEEP = "KEEP"
    CONCATENATE = "CONCATENATE"


class AddEvaluationType(StrEnum):
    KEEP = "KEEP"
    REPLACE = "REPLACE"


class ContinuousTrainingType(StrEnum):
    SCAN = "SCAN"
    EXPERIMENT = "EXPERIMENT"


class ContinuousTrainingTrigger(StrEnum):
    FEEDBACK_LOOP = "FEEDBACK_LOOP"
    NONE = "NONE"


class ContinuousDeploymentPolicy(StrEnum):
    DEPLOY_CHAMPION = "DEPLOY_CHAMPION"
    DEPLOY_SHADOW = "DEPLOY_SHADOW"
    DEPLOY_MANUAL = "DEPLOY_MANUAL"


class ServiceMetrics(str, Enum):
    def __new__(cls, *args, **kwargs):
        value = len(cls.__members__) + 1
        obj = str.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, service, is_aggregation):
        self.service = service
        self.is_aggregation = is_aggregation
        return

    PREDICTIONS_OUTLYING_SCORE = "ae_outlier", False
    PREDICTIONS_DATA = "metrics", False
    REVIEWS_OBJECT_DETECTION_STATS = "object_detection", False
    REVIEWS_CLASSIFICATION_STATS = "classification", False
    REVIEWS_LABEL_DISTRIBUTION_STATS = "label_distribution", False

    AGGREGATED_LABEL_DISTRIBUTION = "label_distribution", True
    AGGREGATED_OBJECT_DETECTION_STATS = "object_detection", True
    AGGREGATED_PREDICTIONS_DATA = "metrics", True
    AGGREGATED_DRIFTING_PREDICTIONS = "ks_drift", True


class ProcessingType(StrEnum):
    PRE_ANNOTATION = "PRE_ANNOTATION"
    DATA_AUGMENTATION = "DATA_AUGMENTATION"
    DATASET_VERSION_CREATION = "DATASET_VERSION_CREATION"
    AUTO_TAGGING = "AUTO_TAGGING"
    AUTO_ANNOTATION = "AUTO_ANNOTATION"


class SupportedContentType(StrEnum):
    PNG = "IMAGE/PNG"
    JPEG = "IMAGE/JPEG"



import logging
from functools import partial
from typing import List, Optional, Union

import orjson
from beartype import beartype
from deprecation import deprecated

import picsellia.utils as utils
from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.model_version import ModelVersion
from picsellia.sdk.tag import Tag, TagTarget
from picsellia.sdk.taggable import Taggable
from picsellia.types.enums import Framework, InferenceType
from picsellia.types.schemas import ModelSchema

logger = logging.getLogger("picsellia")


class Model(Dao, Taggable):
    def __init__(self, connexion: Connexion, data: dict):
        Dao.__init__(self, connexion, data)
        Taggable.__init__(self, TagTarget.MODEL)

    @property
    def name(self) -> str:
        """Name of this (Model)"""
        return self._name

    @property
    @deprecated(
        deprecated_in="6.7.0",
        details="'type' property is deprecated and will be removed in future versions.",
    )
    def type(self) -> InferenceType:
        """Type of this (Model)"""
        return self._type

    @property
    @deprecated(
        deprecated_in="6.7.0",
        details="'framework' property is deprecated and will be removed in future versions.",
    )
    def framework(self) -> Framework:
        """Framework of this (Model)"""
        return self._framework

    @property
    def private(self) -> bool:
        """Privacy of this (Model)"""
        return self._private

    def __str__(self):
        is_private = "" if self._private else "[PUBLIC]"
        return f"{Colors.BLUE}{is_private} Model '{self.name}' {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
            print(foo_dataset.get_resource_url_on_platform())
            >>> "https://app.picsellia.com/model/62cffb84-b92c-450c-bc37-8c4dd4d0f590"
            ```

        Returns:
            Url on Platform for this resource
        """

        return f"{self.connexion.host}/model/{self.id}"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/sdk/model/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict):
        schema = ModelSchema(**data)
        self._name = schema.name
        self._type = schema.type
        self._framework = schema.framework
        self._private = schema.private
        return schema

    @exception_handler
    @beartype
    def update(
        self,
        name: Optional[str] = None,
        framework: Union[str, Framework, None] = None,
        private: Optional[bool] = None,
        description: Optional[str] = None,
        type: Union[str, InferenceType, None] = None,
    ) -> None:
        """Update a model with a new name, framework, privacy, description or type

        Examples:
            ```python
            model.update(description="Very cool model")
            ```
        """
        if type:
            logging.warning(
                "'type' parameter is deprecated and will be removed in future versions. "
                "If you want to give a type to your model version, call update() on ModelVersion."
            )

        if framework:
            logging.warning(
                "'framework' parameter is deprecated and will be removed in future versions. "
                "If you want to give a framework to your model version, call update() on your ModelVersion."
            )

        payload = {
            "name": name,
            "private": private,
            "description": description,
        }
        filtered_payload = utils.filter_payload(payload)
        r = self.connexion.patch(
            f"/sdk/model/{self.id}", data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated.")

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete model.

        Delete the model in Picsellia database

        Examples:
            ```python
            model.delete()
            ```
        """
        self.connexion.delete(f"/sdk/model/{self.id}")
        logger.info(f"{self} deleted.")

    @exception_handler
    @beartype
    def get_tags(self) -> List[Tag]:
        """Retrieve the tags of your model.

        Examples:
            ```python
            tags = my_model.get_tags()
            assert tags[0].name == "my-model-1"
            ```

        Returns:
            List of tags as Tag
        """
        r = self.sync()
        return list(map(partial(Tag, self.connexion), r["tags"]))

    @exception_handler
    @beartype
    def create_version(
        self,
        docker_image_name: Optional[str] = None,
        docker_flags: Optional[List[str]] = None,
        thumb_object_name: Optional[str] = None,
        notebook_link: Optional[str] = None,
        labels: Optional[dict] = None,
        base_parameters: Optional[dict] = None,
        docker_env_variables: Optional[dict] = None,
        name: Optional[str] = None,
        framework: Union[str, Framework, None] = None,
        type: Union[str, InferenceType, None] = None,
    ) -> ModelVersion:
        """Create a version of a model.

        The version number of this model will be defined by the platform. It is incremented automatically.

        Examples:
            ```python
            model_v0 = model.create_version(labels={"1": "cat", "2": "dog"}, framework=Framework.TENSORFLOW, type=InferenceType.OBJECT_DETECTION)
            ```

        Returns:
            A (ModelVersion) object
        """
        if framework:
            framework = Framework.validate(framework)

        if type:
            type = InferenceType.validate(type)

        if type and type not in [
            InferenceType.CLASSIFICATION,
            InferenceType.OBJECT_DETECTION,
            InferenceType.SEGMENTATION,
        ]:
            raise TypeError(f"Type '{type}' not supported yet for model version")

        payload = {
            "name": name,
            "type": type,
            "framework": framework,
            "docker_image_name": docker_image_name,
            "docker_flags": docker_flags,
            "thumb_object_name": thumb_object_name,
            "notebook_link": notebook_link,
            "labels": labels,
            "base_parameters": base_parameters,
            "docker_env_variables": docker_env_variables,
        }
        filtered_payload = utils.filter_payload(payload)
        r = self.connexion.post(
            f"/sdk/model/{self.id}/versions",
            data=orjson.dumps(filtered_payload),
        ).json()
        return ModelVersion(self.connexion, r)

    @exception_handler
    @beartype
    def get_version(self, version: Union[int, str]) -> ModelVersion:
        """Retrieve a version of a model from its version or its name

        Examples:
            ```python
            # Assuming model is a Model without version
            model_version_a = model.create_version("first-version")
            model_version_b = model.get_version(0)
            model_version_c = model.get_version("first-version")
            assert model_version_a == model_version_b
            assert model_version_a == model_version_c
            ```

        Returns:
            A (ModelVersion) object
        """
        if isinstance(version, str):
            params = {"name": version}
        else:
            params = {"version": version}
        r = self.connexion.get(
            f"/sdk/model/{self.id}/versions/find", params=params
        ).json()
        return ModelVersion(self.connexion, r)

    @exception_handler
    @beartype
    def list_versions(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order_by: Optional[List[str]] = None,
    ) -> List[ModelVersion]:
        """List versions of a model

        Examples:
            ```python
            versions = model.list_versions()
            ```

        Returns:
            A list of (ModelVersion) object of this model
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = utils.filter_payload(params)
        r = self.connexion.get(f"/sdk/model/{self.id}/versions", params=params).json()
        return list(map(partial(ModelVersion, self.connexion), r["items"]))



from functools import partial
from typing import List, Optional
from uuid import UUID

from orjson import orjson

from picsellia import exceptions
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.datasource import DataSource
from picsellia.utils import filter_payload


class DataSourceService:
    @classmethod
    def create_datasource(
        cls, connexion: Connexion, organization_id: UUID, name: str
    ) -> DataSource:
        params = {"name": name}
        r = connexion.post(
            f"/sdk/organization/{organization_id}/datasources",
            data=orjson.dumps(params),
        ).json()
        return DataSource(connexion, r)

    @classmethod
    def get_datasource(
        cls, connexion: Connexion, organization_id: UUID, name: str
    ) -> DataSource:
        params = {"name": name}
        r = connexion.get(
            f"/sdk/organization/{organization_id}/datasources/find",
            params=params,
        ).json()
        return DataSource(connexion, r)

    @classmethod
    def get_or_create_datasource(
        cls, connexion: Connexion, organization_id: UUID, name: str
    ) -> DataSource:
        try:
            return cls.get_datasource(connexion, organization_id, name)
        except exceptions.ResourceNotFoundError:
            return cls.create_datasource(connexion, organization_id, name)

    @classmethod
    def list_datasources(
        cls,
        connexion: Connexion,
        organization_id: UUID,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order_by: Optional[List[str]] = None,
    ) -> List[DataSource]:
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = filter_payload(params)
        r = connexion.get(
            f"/sdk/organization/{organization_id}/datasources",
            params=params,
        ).json()
        return list(map(partial(DataSource, connexion), r["items"]))



import logging
import warnings
from functools import partial
from typing import Dict, List, Optional, Union
from uuid import UUID

import orjson
from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.exceptions import BadConfigurationScanError
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.dataset_version import DatasetVersion
from picsellia.sdk.model_version import ModelVersion
from picsellia.sdk.run import Run
from picsellia.sdk.scan_file import ScanFile
from picsellia.types.schemas import ScanSchema
from picsellia.utils import filter_payload, generate_requirements_json

logger = logging.getLogger("picsellia")
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)


class Scan(Dao):
    def __init__(self, connexion: Connexion, data: dict) -> None:
        Dao.__init__(self, connexion, data)

    @property
    def name(self) -> str:
        """Name of this (Scan)"""
        return self._name

    def __str__(self):
        return f"{Colors.BLUE}Scan '{self.name}' {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/sdk/scan/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> ScanSchema:
        schema = ScanSchema(**data)
        self._name = schema.name
        return schema

    @exception_handler
    @beartype
    def update(
        self,
        name: Optional[str] = None,
        metric_name: Optional[str] = None,
        metric_goal: Optional[str] = None,
        requirements: Union[List[Dict], str, None] = None,
        base_model: Optional[ModelVersion] = None,
        base_dataset: Optional[DatasetVersion] = None,
        script: Optional[ScanFile] = None,
        image: Optional[str] = None,
    ) -> None:
        """Update this data source with a new name.

        Examples:
            ```python
            sdk_source.update(name="new name")
            ```
        """
        payload = {
            "name": name,
            "metric_name": metric_name,
            "metric_goal": metric_goal,
            "image": image,
        }
        filtered_payload = filter_payload(payload)
        if base_model is not None:
            filtered_payload["base_model_id"] = base_model.id
        if base_dataset is not None:
            filtered_payload["base_dataset_id"] = base_dataset.id
        if script is not None:
            filtered_payload["script_id"] = script.id
        if isinstance(requirements, str):
            filtered_payload["requirements"] = generate_requirements_json(requirements)
        elif isinstance(requirements, List):
            filtered_payload["requirements"] = requirements
        r = self.connexion.patch(
            f"/sdk/scan/{self.id}", data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def launch(self, gpus: int = 1) -> None:
        """Distribute runs remotely for this scan.

        :information-source: The remote environment has to be setup prior launching the experiment.
        It defaults to our remote training engine.

        Examples:
            ```python
            scan.launch()
            ```
        """
        payload = {
            "gpus": gpus,
        }
        self.connexion.post(f"/sdk/scan/{self.id}/launch", data=orjson.dumps(payload))
        logger.info(f"{self} launched.")

    @exception_handler
    @beartype
    def list_runs(self) -> List[Run]:
        """Retrieve all runs of this scan

        Examples:
            ```python
            scan.list_runs()
            ```

        Returns:
            A list of (Run) object manipulable
        """
        r = self.connexion.get(f"/sdk/scan/{self.id}/runs").json()
        return list(map(partial(Run, self.connexion), r["items"]))

    @exception_handler
    @beartype
    def get_run(self, order: int) -> Run:
        """Retrieve a run object by its order in its scan

        Examples:
            ```python
            scan.get_run(1)
            ```
        Arguments:
            order (int): order of the run

        Returns:
            A (Run) object manipulable
        """
        params = {"order": order}
        r = self.connexion.get(f"/sdk/scan/{self.id}/runs/find", params=params).json()
        return Run(self.connexion, r)

    @exception_handler
    @beartype
    def get_run_by_id(self, id: Union[UUID, str]) -> Run:
        """Retrieve a run object by its id.

        Examples:
            ```python
            scan.get_run_by_id("cb750009-4e09-42bb-8c84-cc78aa004bf0")
            ```
        Arguments:
            id (str): id (primary key) of the run on Picsellia

        Returns:
            A (Run) object manipulable
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(f"/sdk/scan/{self.id}/runs/find", params=params).json()
        return Run(self.connexion, r)

    @exception_handler
    @beartype
    def get_next_run(self) -> Run:
        """Get next available Run for Scan.

        Examples:
            ```python
            scan.get_next_run()
            ```

        Returns:
            A (Run) object manipulable
        """
        r = self.connexion.get(f"/sdk/scan/{self.id}/run/next").json()
        return Run(self.connexion, r)

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this scan from the platform.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            scan.delete()
            ```
        """
        self.connexion.delete(f"/sdk/scan/{self.id}")
        logger.info(f"{self} deleted.")

    @exception_handler
    @beartype
    def get_script(
        self,
    ) -> ScanFile:
        """Retrieve the script of this scan.

        Returns:
            A (ScanFile) object
        """
        r = self.sync()

        if "script_id" not in r or r["script_id"] is None:
            raise BadConfigurationScanError("This scan has no script.")

        r = self.connexion.get(f"/sdk/scan/file/{r['script_id']}").json()
        return ScanFile(self.connexion, r)

    @exception_handler
    @beartype
    def list_data_files(
        self,
    ) -> List[ScanFile]:
        """List all data files of this scan

        Examples:
            ```python
            files = scan.list_data_files()
            ```

        Returns:
            List of (ScanFile) object
        """
        r = self.connexion.get(f"/sdk/scan/{self.id}/scanfiles").json()
        return list(map(partial(ScanFile, self.connexion), r["items"]))



import logging
import os
import warnings
from functools import partial
from pathlib import Path
from typing import List, Optional, Tuple, Union
from uuid import UUID

import orjson
from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from deprecation import deprecated
from PIL import Image

import picsellia.pxl_multithreading as mlt
from picsellia import exceptions
from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.exceptions import (
    NoConnectorFound,
    NoDataError,
    NothingDoneError,
    UploadError,
)
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.data import Data, MultiData
from picsellia.sdk.datasource import DataSource
from picsellia.sdk.tag import Tag
from picsellia.services.datasource import DataSourceService
from picsellia.services.error_manager import ErrorManager
from picsellia.types.enums import DataType, TagTarget
from picsellia.types.schemas import DatalakeSchema
from picsellia.utils import (
    combine_two_ql,
    convert_tag_list_to_query_language,
    filter_payload,
    get_image_shape_with_exif_transpose,
)

logger = logging.getLogger("picsellia")
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)


class Datalake(Dao):
    def __init__(self, connexion: Connexion, organization_id: UUID, data: dict):
        Dao.__init__(self, connexion, data)
        self._organization_id = organization_id

    def __str__(self):
        return f"{Colors.GREEN}Datalake '{self.name}'{Colors.ENDC} (id: {self.id})"

    @property
    def name(self) -> str:
        """Name of this (Datalake)"""
        return self._name

    @property
    def connector_id(self) -> UUID:
        """Connector id used by this (Datalake)"""
        if self._connector_id is None:
            raise NoConnectorFound(
                "This datalake has no connector. You cannot retrieve and upload data into this datalake."
            )
        return self._connector_id

    @exception_handler
    @beartype
    def refresh(self, data: dict):
        schema = DatalakeSchema(**data)
        self._name = schema.name
        if schema.connector_id is not None:
            self._connector_id = schema.connector_id
        return schema

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/sdk/datalake/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
            print(foo_dataset.get_resource_url_on_platform())
            >>> "https://app.picsellia.com/datalake/62cffb84-b92c-450c-bc37-8c4dd4d0f590"
            ```

        Returns:
            Url on Platform for this resource
        """

        return f"{self.connexion.host}/datalake/{self.id}"

    @exception_handler
    @beartype
    def upload_data(
        self,
        filepaths: Union[str, Path, List[Union[str, Path]]],
        tags: Optional[List[Union[str, Tag]]] = None,
        source: Union[str, DataSource, None] = None,
        max_workers: Optional[int] = None,
        error_manager: Optional[ErrorManager] = None,
    ) -> Union[Data, MultiData]:
        """Upload data into this datalake.

        Upload files representing data, into a datalake.
        You can give some tags as a list.
        You can give a source for your data.

        If some data fails to upload, check the example to see how
        to retrieve the list of file paths that failed.

        Examples:
            ```python
            tag_car = client.get_data_tag("car")
            tag_huge_car = client.get_data_tag("huge-car")
            source_camera_one = client.get_datasource("camera-one")
            source_camera_two = client.get_datasource("camera-two")

            lake = client.get_datalake()
            lake.upload_data(filepaths=["porsche.png", "ferrari.png"], tags=[tag_car], source=source_camera_one)
            lake.upload_data(filepaths="truck.png", tags=[tag_huge_car], source=source_camera_two)


            error_manager = ErrorManager()
            lake.upload_data(filepaths=["twingo.png", "path/unknown.png", error_manager=error_manager)

            # This call will return a list of UploadError to see what was wrong
            error_paths = [error.path for error in error_manager.errors]
            ```
        Arguments:
            filepaths (str or Path or List[str or Path]): Filepaths of your data
            tags (List[Tag], optional): Data Tags that will be given to data. Defaults to [].
            source (DataSource, optional): Source of your data.
            max_workers (int, optional): Number of max workers used to upload. Defaults to os.cpu_count() + 4.
            error_manager (ErrorManager, optional): Giving an ErrorManager will allow you to retrieve errors

        Returns:
            A (Data) object or a (MultiData) object that wraps a list of Data.
        """
        computed_tags_ids = []
        if tags:
            for tag in tags:
                if isinstance(tag, str):
                    computed_tag = self.get_or_create_data_tag(tag)
                else:
                    if tag.target_type == TagTarget.DATA:
                        computed_tag = tag
                    else:
                        computed_tag = self.create_data_tag(tag.name)

                computed_tags_ids.append(computed_tag.id)

        if source and isinstance(source, str):
            source = DataSourceService.get_or_create_datasource(
                self.connexion, self._organization_id, source
            )

        def upload_filepath(path: Union[str, Path]):
            try:
                # Retrieve presigned url from platform
                filename = os.path.split(path)[-1]
                object_name = self.connexion.generate_data_object_name(
                    filename, self.connector_id
                )
                _, _, content_type = self.connexion.upload_file(
                    object_name, path, connector_id=self.connector_id
                )

                with Image.open(path) as image:
                    width, height = get_image_shape_with_exif_transpose(image)

                payload = {
                    "type": DataType.IMAGE,
                    "filename": filename,
                    "object_name": object_name,
                    "content_type": content_type,
                    "meta": {
                        "height": height,
                        "width": width,
                    },
                    "tags": computed_tags_ids,
                    "data_source_id": source.id if source else None,
                }

                r = self.connexion.post(
                    f"/sdk/datalake/{self.id}/datas",
                    data=orjson.dumps(payload),
                ).json()
                return Data(self.connexion, self.id, r)
            except Exception as e:
                message = f"Could not upload path '{path}'"
                logger.error(message, exc_info=e)
                if error_manager:
                    error_manager.append(
                        UploadError(message=message, path=path, parent=e)
                    )
                return None

        if isinstance(filepaths, str) or isinstance(filepaths, Path):
            filepaths = [filepaths]

        logger.info("🌎 Starting upload..")

        results = mlt.do_mlt_function(
            filepaths, upload_filepath, max_workers=max_workers
        )
        data_list: List[Data] = [v for _, v in results.items() if v is not None]

        if len(data_list) != len(filepaths):
            logger.error(f"❌ {len(filepaths) - len(data_list)} data not uploaded.")
            if error_manager:
                logger.error(
                    "Calling error_manager.errors will return a list of UploadError objects to see what happened"
                )

        if len(data_list) == 0:
            raise NothingDoneError("Nothing has been uploaded.")
        elif len(data_list) == 1:
            logger.info(f"✅ {data_list[0].filename} data uploaded in {self}")
            return data_list[0]
        else:
            logger.info(f"✅ {len(data_list)} data uploaded in {self}")
            return MultiData(self.connexion, self.id, data_list)

    @exception_handler
    @beartype
    def find_data(
        self,
        filename: Optional[str] = None,
        object_name: Optional[str] = None,
        id: Union[str, UUID, None] = None,
    ) -> Data:
        """Find a data into this datalake

        You can find it by giving its filename or its object name or its id

        Examples:
            ```python
            my_data = my_datalake.find_asset(filename="test.png")
            ```
        Arguments:
            filename (str, optional): filename of the data. Defaults to None.
            object_name (str, optional): object name in the storage S3. Defaults to None.
            id (str or UUID, optional): id of the data. Defaults to None

        Raises:
            If no data match the query, it will raise a NotFoundError.
            In some case, it can raise an InvalidQueryError,
                it might be because platform stores 2 data matching this query (for example if filename is duplicated)

        Returns:
            The (Data) found
        """
        assert not (
            filename is None and object_name is None and id is None
        ), "Select at least one criteria to find a data"

        params = {}
        if id is not None:
            params["id"] = id

        if filename is not None:
            params["filename"] = filename

        if object_name is not None:
            params["object_name"] = object_name

        r = self.connexion.get(
            f"/sdk/datalake/{self.id}/datas/find", params=params
        ).json()
        return Data(self.connexion, self.id, r)

    @exception_handler
    @beartype
    def list_data(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        page_size: Optional[int] = None,
        order_by: Optional[List[str]] = None,
        tags: Union[str, Tag, List[Union[str, Tag]], None] = None,
        filenames: Optional[List[str]] = None,
        intersect_tags: Optional[bool] = False,
        object_names: Optional[List[str]] = None,
        q: Optional[str] = None,
    ) -> MultiData:
        """List data of this datalake.

        If there is no data, raise a NoDataError exception.

        Returned object is a MultiData. An object that allows manipulation of a bunch of data.
        You can add tags on them or feed a dataset with them.

        Examples:
            ```python
            lake = client.get_datalake()
            data = lake.list_data()
            ```

        Args:
            limit (int, optional): if given, will limit the number of data returned
            offset (int, optional): if given, will return data that would have been returned
                                    after this offset in given order
            page_size (int, optional): page size when returning data paginated, can change performance
            order_by (list[str], optional): if not empty, will order data by fields given in this parameter
            filenames (list[str], optional): if given, will return data that have filename equals to one of given filenames
            object_names (list[str], optional): if  given, will return data that have object name equals to one of given object names
            tags (str, (Tag), list[(Tag) or str], optional): if given, will return data that have one of given tags
                                                            by default. if `intersect_tags` is True, it will return data
                                                            that have all the given tags
            intersect_tags (bool, optional): if True, and a list of tags is given, will return data that have
                                             all the given tags. Defaults to False.
            q (str, optional): if given, will filter data with given query. Defaults to None.

        Raises:
            NoDataError: When datalake has no data, raise this exception.

        Returns:
            A (MultiData) object that wraps a list of (Data).
        """
        tags_q = convert_tag_list_to_query_language(tags, intersect_tags)
        query = combine_two_ql(q, tags_q)

        datas = mlt.do_paginate(
            limit,
            offset,
            page_size,
            partial(self._do_list_data, query, order_by, filenames, object_names),
        )

        if len(datas) == 0:
            raise NoDataError("No data found in this datalake with this query")

        return MultiData(self.connexion, self.id, datas)

    @exception_handler
    @beartype
    def _do_list_data(
        self,
        q: Optional[str],
        order_by: Optional[List[str]],
        filenames: Optional[List[str]],
        object_names: Optional[List[str]],
        limit: int,
        offset: int,
    ) -> Tuple[List[Data], int]:
        params = {"limit": limit, "offset": offset}

        if order_by is not None:
            params["order_by"] = order_by
        if q is not None:
            params["q"] = q

        if filenames or object_names:
            # PIC-27 Large filenames and object names list can't be used in GET method
            payload = {}
            if filenames:
                payload["filenames"] = filenames
            if object_names:
                payload["object_names"] = object_names

            r = self.connexion.xget(
                f"/sdk/datalake/{self.id}/datas",
                params=params,
                data=orjson.dumps(payload),
            ).json()
        else:
            r = self.connexion.get(
                f"/sdk/datalake/{self.id}/datas", params=params
            ).json()

        return list(map(partial(Data, self.connexion, self.id), r["items"])), r["count"]

    @exception_handler
    @beartype
    def create_data_tag(self, name: str) -> Tag:
        """Create a data tag used in this datalake

        Examples:
            ```python
            tag_car = lake.create_data_tag("car")
            ```
        Arguments:
            name (str): Name of this tag

        Returns:
            A (Tag) object
        """
        payload = {"name": name}
        r = self.connexion.post(
            f"/sdk/datalake/{self.id}/tags", data=orjson.dumps(payload)
        ).json()
        return Tag(self.connexion, r)

    @exception_handler
    @beartype
    def get_data_tag(self, name: str) -> Tag:
        """Retrieve a data tag used in this datalake.

        Examples:
            ```python
            tag_car = lake.get_data_tag("car")
            ```
        Arguments:
            name (str): Name of the tag you're looking for

        Returns:
            A (Tag) object
        """
        params = {"name": name}
        r = self.connexion.get(
            f"/sdk/datalake/{self.id}/tags/find", params=params
        ).json()
        return Tag(self.connexion, r)

    @exception_handler
    @beartype
    def get_or_create_data_tag(self, name: str) -> Tag:
        """Retrieve a data tag used in this datalake by its name.
        If tag does not exist, create it and return it.

        Examples:
            ```python
            tag = self.get_or_create_data_tag("new_tag")
            ```
        Arguments:
            name (str): Tag to retrieve or create

        Returns:
            A (Tag) object
        """
        try:
            return self.get_data_tag(name)
        except exceptions.ResourceNotFoundError:
            return self.create_data_tag(name)

    @exception_handler
    @beartype
    def list_data_tags(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order_by: Optional[List[str]] = None,
    ) -> List[Tag]:
        """List all tags of this datalake

        Examples:
            ```python
            tags = lake.list_data_tags()
            assert tag_car in tags
            ```

        Returns:
            A List of (Tag)
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = filter_payload(params)
        r = self.connexion.get(f"/sdk/datalake/{self.id}/tags", params=params).json()
        return list(map(partial(Tag, self.connexion), r["items"]))

    @exception_handler
    @beartype
    @deprecated(
        deprecated_in="6.3.2", details="This method can be replaced by list_data()"
    )
    def find_all_datas(self, object_names: List[str]) -> MultiData:
        return self.list_data(object_names=object_names)



from typing import Dict, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field

from picsellia.types.enums import (
    AnnotationStatus,
    DataType,
    ExperimentStatus,
    Framework,
    InferenceType,
    JobRunStatus,
    JobStatus,
    LogType,
    RunStatus,
    TagTarget,
)


class DaoSchema(BaseModel):
    # Might not be defined if refresh is done manually
    id: Optional[UUID]


class DatalakeSchema(DaoSchema):
    name: str
    connector_id: Optional[UUID]


class OrganizationSchema(DaoSchema):
    name: str
    default_connector_id: Optional[UUID]


class DatasetSchema(DaoSchema):
    name: str


class DatasetVersionSchema(DaoSchema):
    name: str = Field(alias="origin_name")
    origin_id: UUID = Field(alias="origin_id")
    version: str
    type: InferenceType


class DatasetVersionStats(BaseModel):
    label_repartition: Dict[str, int]
    nb_objects: int
    nb_annotations: int


class ModelSchema(DaoSchema):
    name: str
    type: InferenceType
    framework: Framework
    private: bool


class ModelVersionSchema(DaoSchema):
    origin: ModelSchema
    name: str
    version: int
    labels: Optional[dict]
    type: InferenceType
    framework: Framework


class ModelDataSchema(DaoSchema):
    name: str
    version_id: UUID
    repartition: dict


class ModelContextSchema(DaoSchema):
    experiment_id: Optional[UUID]
    datas: Optional[List[ModelDataSchema]]
    parameters: dict


class ModelFileSchema(DaoSchema):
    name: str
    object_name: str
    filename: str
    large: bool
    url: Optional[str] = Field(alias="presigned_url")


class ProjectSchema(DaoSchema):
    name: str


class DeploymentSchema(DaoSchema):
    name: str
    type: InferenceType
    oracle_host: Optional[str]
    serving_host: Optional[str]


class ImageMetaSchema(BaseModel):
    width: int
    height: int


class VideoMetaSchema(BaseModel):
    duration: int


class DataSchema(DaoSchema):
    object_name: str
    filename: str
    type: DataType
    url: Optional[str] = Field(alias="presigned_url")


class ImageSchema(DataSchema):
    meta: ImageMetaSchema


class VideoSchema(DataSchema):
    meta: VideoMetaSchema


class DataSourceSchema(DaoSchema):
    name: str


class AssetSchema(DaoSchema):
    data: Union[ImageSchema, VideoSchema]


class ExperimentSchema(DaoSchema):
    name: str
    status: ExperimentStatus


class EvaluationSchema(DaoSchema):
    asset_id: UUID


class UserSchema(DaoSchema):
    username: str


class CollaboratorSchema(DaoSchema):
    user: UserSchema
    organization_id: UUID


class UsernameCollaboratorSchema(DaoSchema):
    username: str


class WorkerSchema(DaoSchema):
    collaborator: UsernameCollaboratorSchema


class ScanSchema(DaoSchema):
    name: str


class RunSchema(DaoSchema):
    order: int
    status: RunStatus
    parameters: dict
    scan_id: UUID
    experiment_id: UUID


class ScanFileSchema(DaoSchema):
    name: str
    object_name: str
    large: bool
    url: Optional[str] = Field(alias="presigned_url")


class ArtifactSchema(DaoSchema):
    name: str
    object_name: str
    filename: str
    large: bool
    url: Optional[str] = Field(alias="presigned_url")


class LoggingFileSchema(DaoSchema):
    object_name: str
    url: Optional[str] = Field(alias="presigned_url")


LogDataType = Union[list, dict, float, int, str]


class LogSchema(DaoSchema):
    name: str
    type: LogType
    data: LogDataType


class TagSchema(DaoSchema):
    name: str
    target_type: TagTarget


class LabelSchema(DaoSchema):
    name: str


class AnnotationSchema(DaoSchema):
    worker_id: UUID
    duration: float
    status: AnnotationStatus


class ShapeSchema(DaoSchema):
    label: LabelSchema


class RectangleSchema(ShapeSchema):
    x: int
    y: int
    w: int
    h: int


class PolygonSchema(ShapeSchema):
    coords: List = Field(alias="polygon")


class LineSchema(ShapeSchema):
    coords: List = Field(alias="line")


class PointSchema(ShapeSchema):
    coords: List = Field(alias="point")
    order: int


class ClassificationSchema(ShapeSchema):
    pass


class ConfigRequirementScan(BaseModel):
    package: str
    version: str


class ExecutionConfigScan(BaseModel):
    type: str


class ConfigScan(BaseModel):
    execution: str
    script: Optional[str]
    requirements: Union[str, List[ConfigRequirementScan], None]
    data: Optional[List[str]]


class ConnectorSchema(DaoSchema):
    bucket_name: str


class JobSchema(DaoSchema):
    status: JobStatus


class JobRunSchema(DaoSchema):
    status: JobRunStatus



import logging
import warnings
from pathlib import Path
from typing import Dict, List, Union

from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from picsellia_annotations.coco import Annotation as COCOAnnotation
from picsellia_annotations.coco import COCOFile
from picsellia_annotations.exceptions import FileError, ParsingError
from picsellia_annotations.utils import read_coco_file

from picsellia.decorators import exception_handler
from picsellia.exceptions import (
    FileNotFoundException,
    UnparsableAnnotationFileException,
)
from picsellia.sdk.label import Label
from picsellia.types.enums import InferenceType

warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger("picsellia")


@exception_handler
@beartype
def parse_coco_file(file_path: Union[Path, str]) -> COCOFile:
    try:
        return read_coco_file(file_path=file_path)
    except FileError:
        raise FileNotFoundException(f"{file_path} not found")
    except ParsingError as e:
        raise UnparsableAnnotationFileException(
            f"Could not parse COCO file {file_path} because : {e}"
        )


@exception_handler
@beartype
def read_annotations(
    cocofile_type: InferenceType,
    coco_annotations: List[COCOAnnotation],
    labels: Dict[int, Label],
    assets: Dict[int, str],
) -> List[Dict]:
    annotations_asset_map: Dict[str, Dict] = {}
    for annotation in coco_annotations:
        try:
            label = labels[annotation.category_id]
        except KeyError:  # pragma: no cover
            logger.error(
                f"category_id {annotation.category_id} not found into retrieved labels"
            )
            continue

        try:
            asset_id = assets[annotation.image_id]
        except KeyError:
            logger.error(
                f"image_id {annotation.image_id} not found into retrieved assets"
            )
            continue

        if asset_id not in annotations_asset_map:
            annotations_asset_map[asset_id] = {
                "rectangles": [],
                "classifications": [],
                "polygons": [],
            }

        if cocofile_type == InferenceType.SEGMENTATION:
            if not annotation.is_rle():
                polygon_coords = annotation.polygon_to_list_coordinates()
            else:
                logger.error(
                    f"annotation_id {annotation.id} is a RLE which is not supported yet"
                )
                continue

            for polygon_coord in polygon_coords:
                annotations_asset_map[asset_id]["polygons"].append(
                    {
                        "polygon": polygon_coord,
                        "label_id": label.id,
                    }
                )
        elif cocofile_type == InferenceType.OBJECT_DETECTION:
            annotations_asset_map[asset_id]["rectangles"].append(
                {
                    "x": int(annotation.bbox[0]),
                    "y": int(annotation.bbox[1]),
                    "w": int(annotation.bbox[2]),
                    "h": int(annotation.bbox[3]),
                    "label_id": label.id,
                }
            )
        elif cocofile_type == InferenceType.CLASSIFICATION:
            annotations_asset_map[asset_id]["classifications"].append(
                {"label_id": label.id}
            )

    return [
        {"asset_id": asset_id, **annotation}
        for asset_id, annotation in annotations_asset_map.items()
    ]



import logging
import warnings
from typing import List, Union
from uuid import UUID

from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning

from picsellia import exceptions
from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.multi_object import MultiObject
from picsellia.types.schemas import EvaluationSchema

logger = logging.getLogger("picsellia")
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)


class Evaluation(Dao):
    def __init__(self, connexion: Connexion, context_id: UUID, data: dict) -> None:
        Dao.__init__(self, connexion, data)
        self._context_id = context_id

    @property
    def asset_id(self) -> UUID:
        """UUID of the (Asset) of this (Evaluation)"""
        return self._asset_id

    @property
    def context_id(self) -> UUID:
        """UUID of the context (it can be an experiment or a model version) of this (Evaluation)"""
        return self._context_id

    def __str__(self):
        return f"{Colors.BLUE}Evaluation of {self.asset_id} from context {self.context_id} {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/sdk/evaluation/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> EvaluationSchema:
        schema = EvaluationSchema(**data)
        self._asset_id = schema.asset_id
        return schema

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this evaluation from the platform.
        All evaluated shapes will be deleted!
        This is a very dangerous move.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            one_evaluation.delete()
            ```
        """
        self.connexion.delete(f"/sdk/evaluation/{self.id}")
        logger.info(f"{self} deleted from platform.")


class MultiEvaluation(MultiObject[Evaluation]):
    @beartype
    def __init__(self, connexion: Connexion, context_id: UUID, items: List[Evaluation]):
        MultiObject.__init__(self, connexion, items)
        self._context_id = context_id

    @property
    def context_id(self) -> UUID:
        return self._context_id

    def __str__(self) -> str:
        return f"{Colors.GREEN}MultiEvaluation for context {self.context_id}{Colors.ENDC}  size: {len(self)}"

    def __getitem__(self, key) -> Union[Evaluation, "MultiEvaluation"]:
        if isinstance(key, slice):
            indices = range(*key.indices(len(self.items)))
            evaluations = [self.items[i] for i in indices]
            return MultiEvaluation(self.connexion, self.context_id, evaluations)
        return self.items[key]

    @beartype
    def __add__(self, other) -> "MultiEvaluation":
        self.assert_same_connexion(other)
        items = self.items.copy()
        self._add_other_items_to_items(other, items)
        return MultiEvaluation(self.connexion, self.context_id, items)

    @beartype
    def __iadd__(self, other) -> "MultiEvaluation":
        self.assert_same_connexion(other)
        self._add_other_items_to_items(other, self.items)
        return self

    def _add_other_items_to_items(self, other, items) -> None:
        if isinstance(other, MultiEvaluation):
            if other.context_id != self.context_id:
                raise exceptions.BadRequestError(
                    "These evaluations does not come from the same context"
                )
            items.extend(other.items.copy())
        elif isinstance(other, Evaluation):
            if other.context_id != self.context_id:
                raise exceptions.BadRequestError(
                    "This evaluation does not come from the same context"
                )
            items.append(other)
        else:
            raise exceptions.BadRequestError("You can't add these two objects")

    def copy(self) -> "MultiEvaluation":
        return MultiEvaluation(self.connexion, self.context_id, self.items.copy())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['picsellia', 'picsellia.sdk', 'picsellia.services', 'picsellia.types']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.2.0,<10.0.0',
 'beartype>=0.11.0,<0.12.0',
 'deprecation>=2.1.0,<3.0.0',
 'orjson>=3.7.11,<4.0.0',
 'picsellia-annotations>=0.6.0,<0.7.0',
 'picsellia-connexion-services>=0.2.0,<0.3.0',
 'pydantic>=1.9.1,<2.0.0',
 'pyyaml>=6.0,<7.0',
 'requests>=2.28.1,<3.0.0',
 'tdqm>=0.0.1,<0.0.2']

setup_kwargs = {
    'name': 'picsellia',
    'version': '6.7.1',
    'description': 'Python SDK package for Picsellia MLOps platform',
    'long_description': '# Picsellia SDK\n\nPicsellia Python SDK is a python library that allows connecting to Picsellia platform.\n\n## Documentation\n\nReference of the SDK can be found at [reference](https://documentation.picsellia.com/reference/client)\n\n## Getting started\nDocumentation can be found at [docs](https://documentation.picsellia.com/docs/getting-started).\nStart by installing the Picsellia python package in your environment.\n```\npip install picsellia\n```\n\nThen, initialize a client\n```python\nfrom picsellia import Client\nclient = Client(api_token=<your api token>)\n```\n\nNow, use it to upload data and create a dataset !\n```python\nlake = client.get_datalake()\nuploaded_data = lake.upload_data(filepaths=["pics/twingo.png", "pics/ferrari.png"], tags=["tag_car"])\n\ndataset = client.create_dataset("cars").create_version("first")\ndataset.add_data(uploaded_data)\n```\n\n## What is Picsellia ?\n\nOur mission is to give you all the necessary tools to relieve the burden of AI projects off of your shoulders. As a data scientist / ML engineer / Researcher, you shouldn\'t have to worry about the following topics :\n\n- [💾 Data Management](https://documentation.picsellia.com/docs/data-management)\n- [📈 Experiment Tracking](https://documentation.picsellia.com/docs/experiment-tracking)\n- [📘 Model Management](https://documentation.picsellia.com/docs/export-an-experiment)\n- [🚀 Model Deployment](https://documentation.picsellia.com/docs/serverless)\n- [👀 Model Monitoring](https://documentation.picsellia.com/docs/monitor-model)\n\nPicsellia is the one-stop place for all the life-cycle of your Computer Vision projects, from ideation to production in a single platform 🚀.\n',
    'author': 'Pierre-Nicolas Tiffreau',
    'author_email': 'pierre-nicolas@picsellia.com',
    'maintainer': 'Pierre-Nicolas Tiffreau',
    'maintainer_email': 'pierre-nicolas@picsellia.com',
    'url': 'https://www.picsellia.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0.0',
}


setup(**setup_kwargs)



import logging
import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Callable, Dict, List, Optional

import tqdm

from picsellia.utils import chunk_list

logger = logging.getLogger("picsellia")

try:
    download_bar_mode = os.environ["PICSELLIA_SDK_DOWNLOAD_BAR_MODE"]
    ascii_bar = True
except KeyError:
    download_bar_mode = "1"
    ascii_bar = False


def do_chunk_called_function(
    items: List[Any], f: Callable, chunk_size: int = 100
) -> List:
    if len(items) < chunk_size:
        return [f(items)]

    results = []
    with tqdm.tqdm(
        total=len(items),
        initial=0,
        ncols=100,
        colour="red",
        ascii=ascii_bar,
    ) as pbar:
        for item_chunked, packet in chunk_list(items, chunk_size):
            results.append(f(item_chunked))
            pbar.update(len(item_chunked))
    return results


def do_mlt_function(
    items: List[Any],
    f: Callable,
    h: Callable = lambda _: _,
    max_workers: Optional[int] = None,
) -> Dict:
    if max_workers is None or max_workers <= 0:
        max_workers = os.cpu_count() + 4

    with tqdm.tqdm(
        total=len(items), ncols=100, colour="green", ascii=ascii_bar
    ) as pbar:
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = {executor.submit(f, item): h(item) for item in items}
            results = {}
            for future in as_completed(futures):
                arg = futures[future]
                try:
                    results[arg] = future.result()
                except Exception as e:
                    logger.error(
                        f"Something went wrong while executing an asynchronous task for item {arg}. Error: {e}"
                    )
                    results[arg] = None
                pbar.update(1)
    return results


PAGINATION_LIMIT = 10000


def do_paginate(
    limit: Optional[int], offset: Optional[int], page_size: Optional[int], f: Callable
):
    if page_size is None or page_size <= 0 or page_size > PAGINATION_LIMIT:
        page_size = PAGINATION_LIMIT
    else:
        page_size = page_size

    if limit is None or limit < 0:
        limit = None
        first_limit = page_size
    else:
        first_limit = min(page_size, limit)

    if offset is None or offset < 0:
        first_offset = 0
    else:
        first_offset = offset

    items, total = f(first_limit, first_offset)

    if len(items) == 0:
        return items

    total_count = limit if limit is not None else total
    k = 1
    if _shall_continue(k, page_size, total, limit, len(items)):
        with tqdm.tqdm(
            total=total_count,
            initial=len(items),
            ncols=100,
            colour="blue",
            ascii=ascii_bar,
        ) as pbar:
            while _shall_continue(k, page_size, total, limit, len(items)):
                k_limit = (
                    page_size if limit is None else min(page_size, limit - len(items))
                )
                k_offset = first_offset + k * page_size
                k_items, _ = f(k_limit, k_offset)
                items += k_items
                k += 1
                pbar.update(len(k_items))

    return items


def _shall_continue(
    k: int, page_size: int, total: int, limit: int, computed: int
) -> bool:
    return k * page_size < total and (limit is None or computed < limit)



from typing import List

from pydantic import BaseModel, root_validator, validator

from picsellia.types.enums import InferenceType


def labels_validator(v: List[int]) -> List[int]:
    if not isinstance(v, List):
        raise TypeError(v)

    if len(v) < 1:
        raise ValueError("There must be at least one id")

    return v


def boxes_validator(v: List[List[int]]) -> List[List[int]]:
    if not isinstance(v, List):
        raise TypeError(v)

    if len(v) < 1:
        raise ValueError("There must be at least one id")

    return v


def scores_validator(v: List[float]) -> List[float]:
    if not isinstance(v, List):
        raise TypeError(v)

    if len(v) < 1:
        raise ValueError("There must be at least one score")

    return v


def masks_validator(v: List[List[List[int]]]) -> List[List[List[int]]]:
    if not isinstance(v, List):
        raise TypeError(v)

    if len(v) < 1:
        raise ValueError("There must be at least one mask")

    return v


class PredictionFormat(BaseModel):
    @property
    def model_type(cls) -> InferenceType:
        raise Exception()


class ClassificationPredictionFormat(PredictionFormat):
    detection_classes: List[int]
    detection_scores: List[float]

    @property
    def model_type(cls) -> InferenceType:
        return InferenceType.CLASSIFICATION


class DetectionPredictionFormat(PredictionFormat):
    detection_classes: List[int]
    detection_boxes: List[List[int]]
    detection_scores: List[float]

    _validate_labels = validator("detection_classes", allow_reuse=True)(
        labels_validator
    )
    _validate_scores = validator("detection_scores", allow_reuse=True)(scores_validator)
    _validate_boxes = validator("detection_boxes", allow_reuse=True)(boxes_validator)

    @property
    def model_type(cls) -> InferenceType:
        return InferenceType.OBJECT_DETECTION

    @root_validator
    def check_sizes(cls, values):
        labels, scores, boxes = (
            values.get("detection_classes"),
            values.get("detection_scores"),
            values.get("detection_boxes"),
        )

        if (
            labels is None
            or scores is None
            or boxes is None is None
            or len(labels) != len(scores)
            or len(boxes) != len(labels)
        ):
            raise ValueError("incoherent lists")

        return values


class SegmentationPredictionFormat(PredictionFormat):
    detection_classes: List[int]
    detection_boxes: List[List[int]]
    detection_scores: List[float]
    detection_masks: List[List[List[int]]]

    _validate_labels = validator("detection_classes", allow_reuse=True)(
        labels_validator
    )
    _validate_boxes = validator("detection_boxes", allow_reuse=True)(boxes_validator)
    _validate_scores = validator("detection_scores", allow_reuse=True)(scores_validator)
    _validate_masks = validator("detection_masks", allow_reuse=True)(masks_validator)

    @property
    def model_type(cls) -> InferenceType:
        return InferenceType.SEGMENTATION

    @root_validator
    def check_sizes(cls, values):
        labels, boxes, scores, masks = (
            values.get("detection_classes"),
            values.get("detection_boxes"),
            values.get("detection_scores"),
            values.get("detection_masks"),
        )

        if (
            labels is None
            or scores is None
            or boxes is None
            or masks is None
            or len(labels) != len(scores)
            or len(boxes) != len(labels)
            or len(masks) != len(labels)
        ):
            raise ValueError("incoherent lists")

        return values



import logging
import warnings
from typing import List, Union

import orjson
from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.multi_object import MultiObject
from picsellia.types.enums import TagTarget
from picsellia.types.schemas import TagSchema

logger = logging.getLogger("picsellia")
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)


class Tag(Dao):
    def __init__(self, connexion: Connexion, data: dict) -> None:
        Dao.__init__(self, connexion, data)

    @property
    def name(self) -> str:
        """Name of this (Tag)"""
        return self._name

    @property
    def target_type(self) -> TagTarget:
        """Target type of this tag, can be :
        DATA, ASSET, MODEL_VERSION, MODEL, DATASET_VERSION, DATASET, DEPLOYMENT, PREDICTED_ASSET"""
        return self._target_type

    def __str__(self):
        return f"{Colors.BLUE}Tag '{self.name}'{Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/sdk/tag/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> TagSchema:
        schema = TagSchema(**data)
        self._name = schema.name
        self._target_type = schema.target_type
        return schema

    @exception_handler
    @beartype
    def update(self, name: str) -> None:
        """Update this tag with a new name.

        Examples:
            ```python
            a_tag.update(name="new name")
            ```
        """
        payload = {"name": name}
        r = self.connexion.patch(
            f"/sdk/tag/{self.id}", data=orjson.dumps(payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this tag from the platform.
        All tagged object will not have this tag anymore.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            tag.delete()
            ```
        """
        self.connexion.delete(f"/sdk/tag/{self.id}")
        logger.info(f"{self} deleted.")

    @exception_handler
    @beartype
    def attach_on(self, targets: Union[Dao, List[Dao], MultiObject[Dao]]) -> None:
        """Attach this tag on a list of target.

        Tag needs to be the same target type as the taggable object.
        For example, if it's a Data Tag, it can only be attached on Data.

        If this is not a good target type, it will not raise any Error, but it will not do anything.

        Examples:
            ```python
            data_tag = datalake.create_data_tag("home")
            some_data = datalake.list_data()
            data_tag.attach_on(some_data)
            ```
        """
        if isinstance(targets, Dao):
            targets = [targets]
        payload = [target.id for target in targets]
        r = self.connexion.post(
            f"/sdk/tag/{self.id}/attach", data=orjson.dumps(payload)
        ).json()
        self.refresh(r["tag"])
        logger.info(f"{self} was attached to {r['count']} object(s)")

    @exception_handler
    @beartype
    def detach_from(self, targets: Union[Dao, List[Dao], MultiObject[Dao]]) -> None:
        """Detach this tag from a list of target.

        Tag needs to be the same target type as the taggable object.
        For example, if it's a Data Tag, it can only be detached from a Data.

        If this is not a good target type, it will not raise any Error, but it will not do anything.

        Examples:
            ```python
            data_tag = datalake.create_data_tag("home")
            some_data = datalake.list_data()
            data_tag.attach_on(some_data)

            data_tag.detach_from(some_data)
            ```
        """
        if isinstance(targets, Dao):
            targets = [targets]
        payload = [target.id for target in targets]
        r = self.connexion.post(
            f"/sdk/tag/{self.id}/detach", data=orjson.dumps(payload)
        ).json()
        self.refresh(r["tag"])
        logger.info(f"{self} was detached from {r['count']} object(s)")
